"""
    cfg_loader
    ~~~~~~~~~~

    A library to load configuration

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

from .loader import BaseConfigLoader, YamlConfigLoader
from .schema import ConfigSchema

__version__ = '0.3.0'

__all__ = [
    'ConfigSchema',
    'BaseConfigLoader',
    'YamlConfigLoader',
]



"""
    docs.conf.py
    ~~~~~~~~~~~~

    Cfg-Loader documentation build configuration file

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
from datetime import datetime

import cfg_loader

sys.path.insert(0, os.path.abspath('..'))

BUILD_DATE = datetime.utcnow()

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sqlalchemy': ('https://docs.sqlalchemy.org/en/latest/', None),
    'marshmallow': ('https://marshmallow.readthedocs.io/en/latest/', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source file names.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
PROJECT_NAME = 'Cfg-Loader'
AUTHOR = 'ConsenSys France'

project = PROJECT_NAME
author = AUTHOR
copyright = '{0}, ConsenSys France'.format(BUILD_DATE.year)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

version = release = cfg_loader.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'venv',
    '*.egg-info',
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
        'donate.html',
    ],
}



"""
    tests.conftest
    ~~~~~~~~~~~~~~

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import os

import pytest

TEST_DIR = os.path.dirname(__file__)
CONFIG_DIR = os.path.join(TEST_DIR, 'config')
BASE_CONFIG_PATH = os.path.join(CONFIG_DIR, 'config.yml')


def set_env_vars(env_vars_to_set):
    """Set environment variable to a value and returns a function to reset environment variables to their old value

    :param env_vars_to_set: List of tuples `(var, new_value)` of environment variables to set
    :type env_vars_to_set: list

    Example:
    >>> import os
    >>> os.environ['VARIABLE'] = 'value'
    >>> reset_env_vars = set_env_vars([('VARIABLE', 'new_value')])
    >>> os.environ['VARIABLE']
    'new_value'
    >>> reset_env_vars()
    >>> os.environ['VARIABLE']
    'value'
    """

    vars = [(var[0], var[1], os.environ.get(var[0])) for var in env_vars_to_set]

    # Set environment variables
    for var, new_value, old_value in vars:
        if new_value is None and old_value is not None:
            del os.environ[var]
        elif new_value is not None:
            os.environ[var] = new_value

    def reset_env_vars():
        set_env_vars([(var[0], var[2]) for var in vars])

    return reset_env_vars


@pytest.fixture
def env_vars():
    yield []


@pytest.fixture
def config_path(env_vars):
    reset_env_vars = set_env_vars(env_vars)

    yield BASE_CONFIG_PATH

    reset_env_vars()



"""
    tests.test_utils
    ~~~~~~~~~~~~~~~~

    Test fields

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import pytest
from cfg_loader.fields import PathValidator, Path
from marshmallow import Schema, ValidationError


def test_path_validator(config_path):
    validator = PathValidator()
    with pytest.raises(ValidationError):
        validator('invalid')
    assert validator(config_path) == config_path


def test_path_field(config_path):
    class SchemaTest(Schema):
        path_field = Path()

    data = {
        'path_field': config_path,
    }
    assert SchemaTest().load(data) == {
        'path_field': config_path,
    }



"""
    tests.test_utils
    ~~~~~~~~~~~~~~~~

    Test utility functions

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import pytest

from cfg_loader.utils import parse_yaml, parse_yaml_file, add_prefix


def test_parse_yaml():
    assert parse_yaml("""
    section:
      subsection:
        - one
        - two
    """) is not None


def test_parse_yaml_file(config_path):
    with pytest.raises(TypeError):
        parse_yaml_file(None)

    with pytest.raises(FileNotFoundError):
        parse_yaml_file('unknown')

    assert parse_yaml_file(config_path)


def test_add_prefix():
    raw_dict = {
        'key1': 'value1',
        'key2': {
            'key3': 'value3'
        },
    }
    assert add_prefix(raw_dict, 'prefix_') == {
        'prefix_key1': 'value1',
        'prefix_key2': {
            'key3': 'value3'
        },
    }



"""
    cfg_loader.schema
    ~~~~~~~~~~~~~~~~~

    Implement marshmallow schema loading

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

from .base import ConfigSchema

__all__ = [
    'ConfigSchema',
]



"""
    cfg_loader.interpolation
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Implement functions to substitute environment variable input data

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

import re
import string

from .exceptions import UnsetRequiredSubstitution, InvalidSubstitution

# Brace formatted syntax separators (c.f. https://docs.docker.com/compose/compose-file/#variable-substitution)
SEPARATOR_DEFAULT_IF_EMPTY = ':-'
SEPARATOR_DEFAULT_IF_UNSET = '-'
SEPARATOR_ERROR_IF_EMPTY = ':?'
SEPARATOR_ERROR_IF_UNSET = '?'


class SubstitutionTemplate(string.Template):
    """Class used to substitute environment variables in a string

    It implements specification from docker-compose environ variable substitution
    (c.f. https://docs.docker.com/compose/compose-file/#variable-substitution)

    Examples with basic substitution

    >>> template = SubstitutionTemplate('${VARIABLE}')
    >>> template.substitute({'VARIABLE': 'value'})
    'value'
    >>> template.substitute({'VARIABLE': ''})
    ''
    >>> template.substitute({})
    Traceback (most recent call last):
    ...
    KeyError: 'VARIABLE'

    Examples with substitution if variable is empty or unset (separator: ":-")

    >>> template = SubstitutionTemplate('${VARIABLE:-default}')
    >>> template.substitute({'VARIABLE': 'value'})
    'value'
    >>> template.substitute({'VARIABLE': ''})
    'default'
    >>> template.substitute({})
    'default'

    Examples with substitution if variable is empty (separator: "-"):

    >>> template = SubstitutionTemplate('${VARIABLE-default}')
    >>> template.substitute({'VARIABLE': 'value'})
    'value'
    >>> template.substitute({'VARIABLE': ''})
    ''
    >>> template.substitute({})
    'default'

    Examples with error raised if variable is unset (separator: "?")

    >>> template = SubstitutionTemplate('${VARIABLE?err}')
    >>> template.substitute({'VARIABLE': 'value'})
    'value'
    >>> template.substitute({'VARIABLE': ''})
    ''
    >>> template.substitute({})
    Traceback (most recent call last):
    ...
    cfg_loader.exceptions.UnsetRequiredSubstitution: err

    Examples with error raised if variable is empty or  unset (separator: ":?")

    >>> template = SubstitutionTemplate('${VARIABLE:?err}')
    >>> template.substitute({'VARIABLE': 'value'})
    'value'
    >>> template.substitute({'VARIABLE': ''})
    Traceback (most recent call last):
    ...
    cfg_loader.exceptions.UnsetRequiredSubstitution: err
    >>> template.substitute({})
    Traceback (most recent call last):
    ...
    cfg_loader.exceptions.UnsetRequiredSubstitution: err
    """

    pattern = r"""
        %(delim)s(?:
            (?P<escaped>%(delim)s) |
            (?P<named>%(id)s)      |
            {(?P<braced>%(bid)s)}  |
            (?P<invalid>)
        )
        """ % {
        'delim': re.escape('$'),
        'id': r'[_a-z][_a-z0-9]*',
        'bid': r'[_a-z][_a-z0-9]*(?:(?P<sep>:?[-?])[^}]*)?',
    }

    def substitute(self, mapping):
        """Substitute values indexed by mapping into `template`

        :param mapping: Mapping containing values to substitute
        :type mapping: dict
        """

        def convert(mo):
            named, braced = mo.group('named') or mo.group('braced'), mo.group('braced')
            if braced is not None:
                sep = mo.group('sep')
                if sep:
                    return process_braced_group(braced, sep, mapping)

            if named is not None:
                val = mapping[named]
                return '%s' % (val,)

            if mo.group('escaped') is not None:  # pragma: no branch
                return self.delimiter

            if mo.group('invalid') is not None:  # pragma: no branch
                raise ValueError('Invalid placeholder: {}'.format(self.template))

        return self.pattern.sub(convert, self.template)


def process_braced_group(braced, sep, mapping):
    """Parse a braced formatted syntax and returns substituted value or raise error

    It implements specification from docker-compose environ variable substitution
    (c.f. https://docs.docker.com/compose/compose-file/#variable-substitution)

    :param braced: Braced formatted syntax to substitute
    :type braced: str
    :param sep: Separator in the braced syntax
    :type sep: str
    :param mapping: Mapping with values to substitute
    :type mapping: dict
    """

    if sep == SEPARATOR_DEFAULT_IF_EMPTY:
        var, _, default = braced.partition(SEPARATOR_DEFAULT_IF_EMPTY)
        return mapping.get(var) or default

    elif sep == SEPARATOR_DEFAULT_IF_UNSET:
        var, _, default = braced.partition(SEPARATOR_DEFAULT_IF_UNSET)
        return mapping.get(var, default)

    elif sep == SEPARATOR_ERROR_IF_EMPTY:
        var, _, err = braced.partition(SEPARATOR_ERROR_IF_EMPTY)
        rv = mapping.get(var)
        if not rv:
            raise UnsetRequiredSubstitution(err)
        return rv

    elif sep == SEPARATOR_ERROR_IF_UNSET:  # pragma: no branch
        var, _, err = braced.partition(SEPARATOR_ERROR_IF_UNSET)
        if var in mapping:
            return mapping.get(var)
        raise UnsetRequiredSubstitution(err)


class Interpolator:
    """Class used to substitute environment variables in complex object

    :param substitution_mapping: Mapping with values to substitute
    :type substitution_mapping: dict

    Example

    >>> interpolator = Interpolator(substitution_mapping={'VARIABLE': 'value'})

    >>> interpolator.interpolate('${VARIABLE} in complex string')
    'value in complex string'

    >>> result = interpolator.interpolate_recursive({'key1': '${VARIABLE}', 'key2': ['element', '${EXTRA-default}']})
    >>> result == {'key1': 'value', 'key2': ['element', 'default']}
    True
    """

    def __init__(self, substitution_mapping=None, substitution_template=SubstitutionTemplate):
        self._substitution_template = substitution_template
        self._substitution_mapping = substitution_mapping or {}

    def interpolate(self, string):
        """Substitute environment variable in a string"""
        try:
            return self._substitution_template(string).substitute(self._substitution_mapping)
        except ValueError as e:
            raise InvalidSubstitution(e)

    def interpolate_recursive(self, obj):
        """Substitute environment variable in an object"""

        if isinstance(obj, str):
            return self.interpolate(obj)

        elif isinstance(obj, dict):
            return {key: self.interpolate_recursive(value) for key, value in obj.items()}

        elif isinstance(obj, list):
            return [self.interpolate_recursive(element) for element in obj]

        return obj



"""
    cfg_loader.fields
    ~~~~~~~~~~~~~~~~~

    Implement marshmallow fields to validate against specific input data

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

import os

from marshmallow import validate, fields


class PathValidator(validate.Validator):
    """Validate a path.

    :param error: Error message to raise in case of a validation error. Can be
        interpolated with `{input}`.
    :type error: str
    """

    default_message = 'Path "{input}" does not exist'

    def __init__(self, error=None):
        self.error = error or self.default_message

    def _format_error(self, value):
        return self.error.format(input=value)

    def __call__(self, value):
        message = self._format_error(value)

        if not os.path.exists(value):
            raise validate.ValidationError(message)

        return value


class Path(fields.String):
    """A validated path field. Validation occurs during both serialization and
    deserialization.

    :param args: The same positional arguments that :class:`~marshmallow.fields.String` receives.
    :param kwargs: The same keyword arguments that :class:`~marshmallow.fields.String` receives.
    """

    default_error_messages = {'invalid_path': 'Path "{input}" does not exist'}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Insert validation into self.validators so that multiple errors can be
        self.validators.append(PathValidator(error=self.error_messages['invalid_path']))


class UnwrapNested(fields.Nested):
    """Nested fields class that will unwrapped at deserialization

    Useful when used with UnwrapSchema

    :param prefix: Optional prefix to add to every key when unwrapping a field
    :type prefix: str
    """

    def __init__(self, *args, prefix='', **kwargs):
        self.prefix = prefix
        super().__init__(*args, **kwargs)



"""
    cfg_loader.schema.base
    ~~~~~~~~~~~~~~~~~~~~~~

    Implement base marshmallow schema to deserialize configuration data

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""
import typing

import marshmallow
from marshmallow import Schema, post_load, INCLUDE, types

from ..exceptions import ValidationError
from ..fields import UnwrapNested
from ..interpolator import SubstitutionTemplate, Interpolator
from ..utils import add_prefix


class InterpolatingSchema(Schema):
    """Schema class that interpolate environ variables from input data

    It implements environment variable substitution following specification from docker-compose
    (c.f. https://docs.docker.com/compose/compose-file/#variable-substitution)

    :param substitution_mapping: Mapping containing values to substitute
    :type substitution: dict
    """

    _interpolator_class = Interpolator
    _substitution_template = SubstitutionTemplate

    def __init__(self, *args, substitution_mapping=None, **kwargs):
        self.substitution_mapping = substitution_mapping or {}
        self.interpolator = self._interpolator_class(substitution_mapping=self.substitution_mapping,
                                                     substitution_template=self._substitution_template)
        super().__init__(*args, **kwargs)

    def load(
        self,
        data: typing.Union[
            typing.Mapping[str, typing.Any],
            typing.Iterable[typing.Mapping[str, typing.Any]],
        ],
        *,
        many: bool = None,
        partial: typing.Union[bool, types.StrSequenceOrSet] = None,
        unknown: str = None
    ):
        """Deserialize a data structure to an object defined by this Schema’s fields

        :param data: Data object to load from
        :type data: dict
        :param many: Whether to deserialize ``data`` as a collection
        :type many: bool
        :param partial: whether to ignore missing fields
        :type partial: bool | tuple
        :param unknown: how to handle unknown fields
        :type unknown: None | EXCLUDE | INCLUDE | RAISE
        :returns: Deserialized data
        :type return: dict
        """
        if self.substitution_mapping:
            # substitute environment variables
            data = self.interpolator.interpolate_recursive(data)

        try:
            return super().load(data, many=many, partial=partial, unknown=unknown)
        except marshmallow.exceptions.ValidationError as e:
            raise ValidationError(e.normalized_messages())


class ExtraFieldsSchema(Schema):
    """Schema class that preserves fields provided in input data but that were omitted in schema fields"""

    class Meta:
        # https://marshmallow.readthedocs.io/en/3.0/quickstart.html#handling-unknown-fields
        unknown = INCLUDE


class UnwrapNestedSchema(Schema):
    """Schema class that can unwrap nested fields"""

    @post_load
    def unwrap_nested_fields(self, data, **kwargs):
        unwrap_nested = {}
        for field, value in self.fields.items():
            if isinstance(value, UnwrapNested):
                unwrap_nested.update(add_prefix(data.pop(field, {}), value.prefix))
        data.update(unwrap_nested)

        return data


class ConfigSchema(InterpolatingSchema, ExtraFieldsSchema, UnwrapNestedSchema):
    """Main schema class for declaring a configuration schema

    It inherits every feature from

    - :class:`InterpolatingSchema`
    - :class:`ExtraFieldsSchema`
    - :class:`UnwrapNestedSchema`

    Example

    >>> from cfg_loader import ConfigSchema, BaseConfigLoader
    >>> from marshmallow import fields

    >>> class MyConfigSchema(ConfigSchema):
    ...     setting1 = fields.Str()
    ...     setting2 = fields.Int(required=True)
    ...     setting3 = fields.Float(missing=13.2)

    >>> substitution_mapping = {'VARIABLE': 'substitution'}
    >>> my_config_loader = BaseConfigLoader(MyConfigSchema,
    ...                                     substitution_mapping)

    >>> schema = MyConfigSchema(substitution_mapping=substitution_mapping)
    >>> config = schema.load({
    ...    'setting1': '${VARIABLE}',
    ...    'setting2': '${UNSET_VARIABLE:-1}',
    ... })

    >>> config == {
    ...     'setting1': 'substitution',
    ...     'setting2': 1,
    ...     'setting3': 13.2,
    ... }
    True
    """



"""
    cfg_loader.loader
    ~~~~~~~~~~~~~~~~~

    Implement the loading class

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

import os

from .exceptions import ConfigFileMissingError, ConfigFileNotFoundError
from .utils import parse_yaml_file

# Default environment variable containing the path of a .yaml configuration file
DEFAULT_CONFIG_FILE_ENV_VAR = 'CONFIG_FILE'


class BaseConfigLoader:
    """Base config loader using a marshmallow schema to validate and process input data

    :param schema: Marshmallow schema used to deserialize configuration input data
    """

    def __init__(self, config_schema, substitution_mapping=None):
        self.substitution_mapping = substitution_mapping or {}
        self.config_schema = config_schema

    def load(self, data, substitution_mapping=None):
        """Load configuration from an object

        :param input: Data to load configuration from (must be deserializable by schema)
        :type input: object
        """
        substitution_mapping = substitution_mapping or self.substitution_mapping

        return self.config_schema(substitution_mapping=substitution_mapping).load(data)


class YamlConfigLoader(BaseConfigLoader):
    """Config loader that reads config from .yaml file

    :param config_file_env_var: Environment variable to read config file path from if not provided when loading
    :type config_file_env_var: str
    :param default_config_path: Used if neither path is provided at loading nor environment variable
    :type default_config_path: str
    """

    def __init__(self, *args, config_file_env_var=DEFAULT_CONFIG_FILE_ENV_VAR, default_config_path=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.config_file_env_var = config_file_env_var
        self.default_config_path = default_config_path

    def check_file(self, config_file):
        """Check file validity

        :param config_file: Path to the .yaml configuration file
        :type config_file: str
        """

        if config_file is None:
            raise ConfigFileMissingError("""
                    No configuration file specified.
                    Please provide a configuration valid file path or
                    set the environ variable '{var}'
                """.format(var=self.config_file_env_var))

        if not os.path.isfile(config_file):
            raise ConfigFileNotFoundError("""
                    No such file '{path}'
                """.format(path=config_file))

        return config_file

    def load(self, config_file=None, substitution_mapping=None):
        """Load configuration from .yaml file

        :param config_file: Path to the .yaml configuration file
        :type config_file: str
        """

        config_file = config_file or os.environ.get(self.config_file_env_var) or self.default_config_path

        # Check config_file is valid
        self.check_file(config_file)

        # Parse yaml file
        data = parse_yaml_file(config_file)

        return super().load(data, substitution_mapping)



"""
    tests.test_substitution
    ~~~~~~~~~~~~~~~~~~~~~~~

    Test substitution functions

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import pytest

from cfg_loader.exceptions import UnsetRequiredSubstitution, InvalidSubstitution
from cfg_loader.interpolator import SubstitutionTemplate, Interpolator


@pytest.fixture(scope='module')
def interpolator():
    yield Interpolator(substitution_mapping={'VARIABLE': 'value'})


def _test_substitution_base():
    template = SubstitutionTemplate('${VARIABLE}')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    assert template.substitute({'VARIABLE': ''}) == ''
    with pytest.raises(KeyError):
        template.substitute({})

    template = SubstitutionTemplate('$VARIABLE')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    assert template.substitute({'VARIABLE': ''}) == ''
    with pytest.raises(KeyError):
        template.substitute({})


def _test_substitution_default_if_empty():
    template = SubstitutionTemplate('${VARIABLE:-default}')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    assert template.substitute({'VARIABLE': ''}) == 'default'
    assert template.substitute({}) == 'default'


def _test_substitution_default_if_unset():
    template = SubstitutionTemplate('${VARIABLE-default}')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    assert template.substitute({'VARIABLE': ''}) == ''
    assert template.substitute({}) == 'default'


def _test_substitution_error_if_unset():
    template = SubstitutionTemplate('${VARIABLE?test error}')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    assert template.substitute({'VARIABLE': ''}) == ''
    with pytest.raises(UnsetRequiredSubstitution) as e:
        template.substitute({})
    assert str(e.value) == 'test error'


def _test_substitution_error_if_empty():
    template = SubstitutionTemplate('${VARIABLE:?test error}')

    assert template.substitute({'VARIABLE': 'value'}) == 'value'
    with pytest.raises(UnsetRequiredSubstitution) as e:
        template.substitute({'VARIABLE': ''})
    assert str(e.value) == 'test error'

    with pytest.raises(UnsetRequiredSubstitution) as e:
        template.substitute({})
    assert str(e.value) == 'test error'


def _test_substitution_complex():
    template = SubstitutionTemplate('/usr/${VARIABLE-default}/app')
    assert template.substitute({'VARIABLE': 'src'}) == '/usr/src/app'
    assert template.substitute({}) == '/usr/default/app'

    template = SubstitutionTemplate('\"Hello, ${VARIABLE}!\"')
    assert template.substitute({'VARIABLE': 'world'}) == '\"Hello, world!\"'

    template = SubstitutionTemplate('$$VARIABLE')
    assert template.substitute({'VARIABLE': 'world'}) == '$VARIABLE'


def _test_substitution_ignored():
    template = SubstitutionTemplate('no substitution pattern')
    assert template.substitute({}) == 'no substitution pattern'


def _test_substitution_invalid():
    template = SubstitutionTemplate('${VARIABLE')
    with pytest.raises(ValueError):
        template.substitute({'VARIABLE': 'world'})

    template = SubstitutionTemplate('${VARIABLE }')
    with pytest.raises(ValueError):
        template.substitute({'VARIABLE': 'world'})

    template = SubstitutionTemplate('${VARIABLE!}')
    with pytest.raises(ValueError):
        template.substitute({'VARIABLE': 'world'})

    template = SubstitutionTemplate('${}')
    with pytest.raises(ValueError):
        print(template.substitute({'VARIABLE': 'world'}))


def test_substitution():
    _test_substitution_base()
    _test_substitution_default_if_empty()
    _test_substitution_default_if_unset()
    _test_substitution_error_if_unset()
    _test_substitution_error_if_empty()
    _test_substitution_complex()
    _test_substitution_invalid()


def _test_valid_interpolation(interpolator):
    raw_input = {
        'key1': '${VARIABLE}',
        'key2': [
            'element',
            '${EXTRA-default}',
        ],
    }
    assert interpolator.interpolate_recursive(raw_input) == {
        'key1': 'value',
        'key2': [
            'element',
            'default',
        ],
    }


def _test_invalid_interpolation(interpolator):
    input = '${VARIABLE?test error'

    with pytest.raises(InvalidSubstitution):
        interpolator.interpolate(input)


def test_interpolator(interpolator):
    _test_valid_interpolation(interpolator)
    _test_invalid_interpolation(interpolator)



"""
    tests.test_loader
    ~~~~~~~~~~~~~~~~~

    Test configuration loader

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

import pytest
from marshmallow import fields

from cfg_loader.exceptions import ConfigFileMissingError, ConfigFileNotFoundError
from cfg_loader.loader import BaseConfigLoader, YamlConfigLoader
from cfg_loader.schema import ConfigSchema
from .conftest import BASE_CONFIG_PATH


class BaseConfigSchema(ConfigSchema):
    name = fields.Str()
    path = fields.Str()


class SecurityConfigSchema(ConfigSchema):
    secret = fields.Str()


class ConfigSchemaTest(ConfigSchema):
    base = fields.Nested(BaseConfigSchema)
    security = fields.Nested(SecurityConfigSchema)


def test_base_config_loader():
    config_loader = BaseConfigLoader(ConfigSchemaTest)
    assert config_loader.load({
        'base': {
            'name': 'App-Name',
            'path': '/home/folder',
        },
        'security': {
            'secret': 'my-secret',
        },
    }) == {
        'base': {
            'name': 'App-Name',
            'path': '/home/folder',
        },
        'security': {
            'secret': 'my-secret',
        },
    }


def test_yaml_config_loader(config_path):
    config_loader = YamlConfigLoader(ConfigSchemaTest,
                                     substitution_mapping={'PATH': 'folder/file', 'SECRET': 'my-secret'})

    assert config_loader.load(config_path) == {
        'base': {
            'name': 'App-Name',
            'path': '/home/user/folder/file',
        },
        'security': {
            'secret': 'my-secret',
        },
    }


@pytest.mark.parametrize('env_vars', [[('VARIABLE', BASE_CONFIG_PATH)]])
def test_yaml_config_loader_from_env_var(config_path):
    config_loader = YamlConfigLoader(ConfigSchemaTest,
                                     substitution_mapping={'PATH': 'folder/file', 'SECRET': 'my-secret'},
                                     config_file_env_var='VARIABLE')

    assert config_loader.load() == {
        'base': {
            'name': 'App-Name',
            'path': '/home/user/folder/file',
        },
        'security': {
            'secret': 'my-secret',
        },
    }


@pytest.mark.parametrize('env_vars', [[('VARIABLE', None)]])
def test_yaml_config_loader_no_path(config_path):
    config_loader = YamlConfigLoader(ConfigSchemaTest,
                                     substitution_mapping={'PATH': 'folder/file', 'SECRET': 'my-secret'},
                                     config_file_env_var='VARIABLE')

    with pytest.raises(ConfigFileMissingError) as e:
        config_loader.load()
    assert 'VARIABLE' in str(e.value)


@pytest.mark.parametrize('env_vars', [[('VARIABLE', None)]])
def test_yaml_config_loader_with_default_path(config_path):
    config_loader = YamlConfigLoader(ConfigSchemaTest,
                                     substitution_mapping={'PATH': 'folder/file', 'SECRET': 'my-secret'},
                                     config_file_env_var='VARIABLE',
                                     default_config_path=BASE_CONFIG_PATH)

    assert config_loader.load() == {
        'base': {
            'name': 'App-Name',
            'path': '/home/user/folder/file',
        },
        'security': {
            'secret': 'my-secret',
        },
    }


def test_yaml_config_loader_invalid_path():
    config_loader = YamlConfigLoader(ConfigSchemaTest,
                                     substitution_mapping={'PATH': 'folder/file', 'SECRET': 'my-secret'}, )

    with pytest.raises(ConfigFileNotFoundError) as e:
        config_loader.load('unknown/config/file')
    assert 'unknown/config/file' in str(e.value)



"""
    cfg_loader.utils
    ~~~~~~~~~~~~~~~~

    Implement utilities functions

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""

import yaml


def parse_yaml(content):
    return yaml.safe_load(content)


def parse_yaml_file(path):
    with open(path, 'rt') as f:
        return parse_yaml(f.read())


def add_prefix(dictionary, prefix):
    """Add prefix to every key in a dictionary

    :param dictionary: Original dictionary
    :type dictionary: dict
    :param prefix: Prefix to add to every key
    :type prefix: str
    """
    return {'{}{}'.format(prefix, key): value for key, value in dictionary.items()}



"""
    cfg_loader.exceptions
    ~~~~~~~~~~~~~~~~~~~~~

    Implement configuration loader Exception

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see :ref:`license` for more details.
"""


class ConfigLoaderError(Exception):
    """Base class for all application related errors."""


class ConfigFileMissingError(ConfigLoaderError):
    """Error raised when not specifying a config file"""


class ConfigFileNotFoundError(ConfigLoaderError):
    """Error raised when an invalid config file path is provided"""


class LoadingError(ConfigLoaderError):
    """Error raised when loading the configuration file"""


class ValidationError(ConfigLoaderError):
    """Error raised when marshmallow raise a validation error at deserialization"""


class UnsetRequiredSubstitution(LoadingError):
    """Error raised when a substitution environment variable is required but unset"""


class InvalidSubstitution(LoadingError):
    """Error raised when an invalid substitution is detected"""



"""
    Cfg-Loader
    ~~~~~~~~~~

    Config-Loader is an empty Python project.

    It is meant to be forked when starting a new Python project to inherit structure.

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import os
import re

from setuptools import setup, find_packages


def read(file_name):
    try:
        return open(os.path.join(os.path.dirname(__file__), file_name)).read()
    except FileNotFoundError:
        return ''


def find_version(file):
    """Attempts to find the version number in a file.

    Raises RuntimeError if not found.

    :param file: File where to find version
    :type file: str
    """
    version = ''
    with open(file, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


setup(
    name='theatro-cfg-loader',
    version=find_version('cfg_loader/__init__.py'),
    license="BSD-3-Clause",
    url='https://github.com/nmvalera/cfg-loader',
    author='Nicolas Maurice',
    author_email='nicolas.maurice.valera@gmail.com',
    maintainer='Jacob Floyd',
    maintainer_email='cognifloyd@gmail.com',
    description='A library that allows to easily load configuration settings.',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    install_requires=[
        'marshmallow>=3.0.0',
        'PyYAML>=3.12',
    ],
    extras_require={
        'dev': [
            'flake8',
            'autoflake',
            'autopep8',
            'coverage',
            'pytest>=3',
            'tox',
            'sphinx',
            'sphinx_rtd_theme',
        ],
        'doc': [
            'sphinx',
            'sphinx_rtd_theme',
        ],
    },
    zip_safe=False,
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='tests'
)



"""
    tests.test_schema
    ~~~~~~~~~~~~~~~~~

    Test schema

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import pytest
from marshmallow import fields

from cfg_loader.exceptions import ValidationError
from cfg_loader.fields import UnwrapNested
from cfg_loader.schema import ConfigSchema


class NestedConfigSchemaTest(ConfigSchema):
    field = fields.Int()
    many = fields.List(fields.Str())


class ConfigSchemaTest(ConfigSchema):
    field = fields.Str()
    nested = fields.Nested(NestedConfigSchemaTest)


@pytest.fixture(scope='module')
def basic_config_schema_loader():
    yield ConfigSchemaTest()


@pytest.fixture(scope='module')
def interpolating_config_schema_loader():
    yield ConfigSchemaTest(substitution_mapping={'VARIABLE': 'substitution', 'VARIABLE_INT': '24'})


def _test_invalid_basic_config_schema_loading(schema_loader):
    invalid_raw_config = {
        'field': 2,
        'extra': 'extra_value',
        'nested': {
            'field': 4,
            'many': [
                'element',
            ],
        }
    }
    with pytest.raises(ValidationError):
        schema_loader.load(invalid_raw_config)


def _test_valid_basic_config_schema_loading(schema_loader):
    valid_raw_config = {
        'field': 'value',
        'extra': 'extra_value',
        'nested': {
            'field': '4',
            'many': [
                'element',
            ],
            'extra': {
                'key': 2,
                'many': [
                    'element',
                ],
            }
        }
    }
    assert schema_loader.load(valid_raw_config) == {
        'field': 'value',
        'extra': 'extra_value',
        'nested': {
            'field': 4,
            'many': [
                'element',
            ],
            'extra': {
                'key': 2,
                'many': [
                    'element',
                ],
            },
        },
    }


def test_basic_config_schema_loading(basic_config_schema_loader, interpolating_config_schema_loader):
    for config_loader in [basic_config_schema_loader, interpolating_config_schema_loader]:
        _test_invalid_basic_config_schema_loading(config_loader)
        _test_valid_basic_config_schema_loading(config_loader)


def test_interpolating_config_schema_loading(interpolating_config_schema_loader):
    raw_config = {
        'field': 'value',
        'extra': '${VARIABLE?err}',
        'nested': {
            'field': '${VARIABLE_INT}',
            'many': [
                '${UNSET_VARIABLE-default}',
                'element',
            ],
            'extra': [
                'element',
                {
                    'key': 'value',
                }
            ]
        },
    }
    assert interpolating_config_schema_loader.load(raw_config) == {
        'field': 'value',
        'extra': 'substitution',
        'nested': {
            'field': 24,
            'many': [
                'default',
                'element',
            ],
            'extra': [
                'element',
                {
                    'key': 'value',
                },
            ],
        },
    }


class UnwrapNestedConfigSchemaTest(ConfigSchema):
    nested = UnwrapNested(NestedConfigSchemaTest)


class UnwrapConfigSchemaTest(ConfigSchema):
    field = fields.Str()
    unwrap_nested = UnwrapNested(UnwrapNestedConfigSchemaTest, prefix='unwrapped_')
    regular_nested = fields.Nested(NestedConfigSchemaTest)


@pytest.fixture(scope='module')
def unwrap_config_schema_loader():
    yield UnwrapConfigSchemaTest(substitution_mapping={'VARIABLE': 'substitution', 'VARIABLE_INT': '24'})


def test_unwrap_config_schema_loading(unwrap_config_schema_loader):
    raw_config = {
        'field': 'value',
        'extra': '${VARIABLE?err}',
        'unwrap_nested': {
            'nested': {
                'field': '${VARIABLE_INT}',
                'many': [
                    '${UNSET_VARIABLE-default}',
                    'element',
                ],
            },
            'extra': [
                'element',
            ],
        },
        'regular_nested': {
            'field': 33,
            'many': [
                '${UNSET_VARIABLE-default}',
            ],
        }
    }

    assert unwrap_config_schema_loader.load(raw_config) == {
        'field': 'value',
        'extra': 'substitution',
        'unwrapped_field': 24,
        'unwrapped_many': [
            'default',
            'element',
        ],
        'unwrapped_extra': [
            'element'
        ],
        'regular_nested': {
            'field': 33,
            'many': [
                'default',
            ],
        }
    }


class NestedDataKeyConfigSchemaTest(ConfigSchema):
    nested = fields.Nested(NestedConfigSchemaTest, data_key='nested-key')


def test_nested_data_key_loading():
    schema = NestedDataKeyConfigSchemaTest()
    raw_data = {
        'nested-key': {
            'field': '4',
            'many': [],
        },
    }
    loaded_data = schema.load(raw_data)
    assert loaded_data == {
        'nested': {
            'field': 4,
            'many': [],
        },
    }


class UnwrapNestedDataKeyConfigSchemaTest(ConfigSchema):
    nested = UnwrapNested(NestedConfigSchemaTest, data_key='nested-key')


def test_unwrap_nested_data_key_loading():
    schema = UnwrapNestedDataKeyConfigSchemaTest()
    raw_data = {
        'nested-key': {
            'field': '4',
            'many': [],
        },
    }
    loaded_data = schema.load(raw_data)
    assert loaded_data == {
        'field': 4,
        'many': [],
    }



"""
    tests.test_import
    ~~~~~~~~~~~~~~~~~

    Test import

    :copyright: Copyright 2017 by ConsenSys France.
    :license: BSD, see LICENSE for more details.
"""

import cfg_loader


def _test_version(module):
    assert hasattr(module, '__version__')


def _test_all(module):
    assert hasattr(module, '__all__')


def test_import():
    _test_version(cfg_loader)
    _test_all(cfg_loader)
